# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Utilities to be used."""

import json
import math
import logging
import datetime
from http import HTTPStatus
from pyswitcheo.errors import HTTPResponseError
from pyswitcheo.internal.api import exchange


logger = logging.getLogger(__name__)


def format_urls(base_url, end_point):
    """Create a url given a base url and an end point.

    Args:
        base_url (str): A base url which needs to be queried.
        end_point(str): The other part of the end point
    Returns:
        A properly formatted url to query
    """
    url = '/'.join([str(base_url).strip('/'), str(end_point).strip('/')])
    logger.debug("Url to query: {0}".format(url))
    return url


def convert_to_neo_asset_amount(amount, asset_id, base_url):
    """Convert a given input to a neo asset precision.

    Internally this API queries the Switcheo exchange to get the correct precision
    for a given asset.

    Args:
        amount (int)   : Input amount for which precision needs to be calculated.
        asset_id (str) : A string representing the correct asset.
        base_url (str) : URL of the switcheo exchange which will return the correct decimal precision.
    """
    # sanitize the asset_id
    asset_id = str(asset_id).lower()
    response = exchange._get_contract_tokens_info(base_url)
    json_resp = response_to_json(response)

    # sanitize the json_response, i.e. all keys should be lower
    json_resp_sanitized = {str(k).lower(): v for k, v in json_resp.items()}
    precision = json_resp_sanitized[asset_id]["decimals"]
    return int(amount * math.pow(10, precision))


def response_else_exception(response):
    """Return a response in case HTTPStatus is 200, else raise a custom exception.
    """
    if response.status_code != HTTPStatus.OK:
        logger.debug("Response returned from the server: {0}".format(response.text))
        raise HTTPResponseError(response)
    return response


def response_to_json(response):
    """A simple wrapper to convert requests.content or requests.text to a corresponding json object."""
    logger.debug("Original response {}".format(response.content.decode("UTF-8")))
    json_response = json.loads(response.content.decode("UTF-8"))
    logger.debug("Response dumped as json {0}".format(json_response))
    return json_response


def jsonify(json_obj):
    """Convert a given json object to string with sorted key and without spaces."""
    logger.debug("Json before invoking jsonify {}".format(json_obj))
    return json.dumps(json_obj, sort_keys=True, separators=(",", ":"))


def get_current_epoch_milli():
    """Get the current time in epoch milliseconds."""
    return int(round(datetime.datetime.now().timestamp() * 1000))



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""This module can be used to validate incoming response from Switcheo API server."""


LIST_TRADES_SCHEMA = {
    "type": "array",
    "properties": {
        "id": {"type": "string"},
        "fill_amount": {"type": "number"},
        "take_amount": {"type": "number"},
        "is_buy": {"type": "boolean"},
    },
}

# Sample schema for the response expected from execute_withdrawal end point
EXECUTE_WITHDRAWAL_SCHEMA = {
    "type": "object",
    "properties": {
        "event_type": {"type": "string"},
        "amount": {"type": "number"},
        "asset_id": {"type": "string"},
        "status": {"type": "string"},
        "id": {"type": "string"},
        "blockchain": {"type": "string"},
        "reason_code": {"type": "number"},
        "address": {"type": "string"},
        "transaction_hash": {"type": "object"},
        "created_at": {"type": "string"},
        "updated_at": {"type": "string"},
        "contract_hash": {"type": "string"},
    },
}
# Sample schema for the response expected from create_withdrawal end point
CREATE_WITHDRAWAL_SCHEMA = {"type": "object", "properties": {"id": {"type": "string"}}}

# Sample schema for the response expected from create_deposit end point
CREATE_DEPOSIT_SCHEMA = {
    "type": "object",
    "properties": {
        "id": {"type": "string"},
        "transaction": {
            "type": "object",
            "properties": {
                "hash": {"type": "string"},
                "sha256": {"type": "string"},
                "type": {"type": "number"},
                "version": {"type": "number"},
                "attributes": {"type": "array"},
                "inputs": {"type": "array"},
                "outputs": {"type": "array"},
                "scripts": {"type": "array"},
                "script": {"type": "string"},
                "gas": {"type": "number"},
            },
        },
        "script_params": {
            "type": "object",
            "properties": {
                "scriptHash": {"type": "string"},
                "operation": {"type": "string"},
                "args": {"type": "array"},
            },
        },
    },
}


# This is the response received after creating an order
CREATE_ORDER_RESPONSE_SCHEMA = {
    "type": "object",
    "properties": {
        "id": {"type": "string"},
        "blockchain": {"type": "string"},
        "contract_hash": {"type": "string"},
        "address": {"type": "string"},
        "side": {"type": "string"},
        "offer_asset_id": {"type": "string"},
        "want_asset_id": {"type": "string"},
        "offer_amount": {"type": "string"},
        "want_amount": {"type": "string"},
        "transfer_amount": {"type": "string"},
        "priority_gas_amount": {"type": "string"},
        "use_native_token": {"type": "boolean"},
        "native_fee_transfer_amount": {"type": "number"},
        "deposit_txn": {"type": ["object", "null"]},
        "created_at": {"type": "string"},
        "status": {"type": "string"},
        "fills": {"type": "array"},
        "makes": {"type": "array"},
    },
}



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Base API implementation for pyswitcheo."""

from pyswitcheo.internal.api import balances
from pyswitcheo.internal.api import deposits
from pyswitcheo.internal.api import offers
from pyswitcheo.internal.api import exchange
from pyswitcheo.internal.api import tickers
from pyswitcheo.internal.api import trades
from pyswitcheo.internal.api import orders
from pyswitcheo.internal.api import withdrawals
from pyswitcheo.utils import response_to_json


class SwitcheoApi(object):
    """Base implementation for interacting with pyswitcheo APIs."""

    def __init__(self, base_url, api_version="v2"):
        """Initialize Api class instances.
        Args:
            base_url(str)    : Base url represents the endpoint to query the Switcheo API server.
            api_version(str) : An optional api version which could change in future
        """
        self.base_url = str(base_url).strip("/") + '/' + api_version.strip("/")

    def get_candle_sticks(self, pair, start_time, end_time, interval):
        """Get candlestick chart data filtered by url parameters.

        Args:
            pair (str)	       : Show chart data of this trading pair
            start_time (int)   : Start of time range for data in epoch seconds
            end_time (int)	   : End of time range for data in epoch seconds
            interval (int)	   : Candlestick period in minutes Possible values are: 1, 5, 30, 60, 360, 1440

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response:
            [
              {
                "time": "1531215240",
                "open": "0.00049408",
                "close": "0.00049238",
                "high": "0.000497",
                "low": "0.00048919",
                "volume": "110169445.0",
                "quote_volume": "222900002152.0"
              },
              {
                "time": "1531219800",
                "open": "0.00050366",
                "close": "0.00049408",
                "high": "0.00050366",
                "low": "0.00049408",
                "volume": "102398958.0",
                "quote_volume": "205800003323.0"
              },
              ...
            ]
        """
        return tickers._get_candle_sticks(
            base_url=self.base_url,
            pair=pair,
            start_time=start_time,
            end_time=end_time,
            interval=interval
        )

    def list_contracts(self):
        """Fetch updated hashes of contracts deployed by Switcheo.

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response
            {
              "NEO":
              {
                "V1": "0ec5712e0f7c63e4b0fea31029a28cea5e9d551f",
                "V1_5": "c41d8b0c30252ce7e8b6d95e9ce13fdd68d2a5a8",
                "V2": "48756743d524af03aa75729e911651ffd3cbe7d8"
              }
            }
        """
        return exchange._list_contracts(self.base_url)

    def list_pairs(self, bases):
        """Fetch available currency pairs on Switcheo Exchange filtered by the base parameter. Defaults to all pairs.

        Args:
            base_url (str)    : This paramter governs whether to connect to test or mainnet..
            bases (list[str]) : Provides pairs for these base symbols. Possible values are NEO, GAS, SWTH, USD.

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response
            [
              "GAS_NEO",
              "SWTH_NEO"
            ]
        """
        exchange._list_currency_pairs(self.base_url, bases)

    def get_exchange_timestamp(self):
        """Returns the current timestamp in the exchange.

        This value should be fetched and used when a timestamp parameter is required for API requests.
        If the timestamp used for your API request is not within an acceptable range of the exchange's
        timestamp then an invalid signature error will be returned. The acceptable range might vary, but it
        should be less than one minute.

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response
            {
              "timestamp": 1534392760908
            }
        """
        return exchange._get_exchange_timestamp(self.base_url)

    def get_contract_tokens_info(self):
        """Fetch updated hashes of contracts deployed by Switcheo along with their precision.

        Args:
            base_url (str): This paramter governs whether to connect to test or mainnet.

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response
            {
              "NEO": {
                "hash": "c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b",
                "decimals": 8
              },
              "GAS": {
                "hash": "602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7",
                "decimals": 8
              },
              "SWTH": {
                "hash": "ab38352559b8b203bde5fddfa0b07d8b2525e132",
                "decimals": 8
              },
              ...
            }
        """
        return exchange._get_contract_tokens_info(self.base_url)

    def list_balances(self, addresses, contract_hashes):
        """List contract balances of the given address and contract hashes.

        Args:
            base_url (str)               : This paramter governs whether to connect to test or mainnet..
            addresses (list[str])        : Only return balances for these addresses.
            contract_hashes (list[str])  : Only return balances from these contract hashes.

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response:
            {
              "confirming": {
                "GAS": [
                  {
                    "event_type": "withdrawal",
                    "asset_id": "602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7",
                    "amount": -100000000,
                    "transaction_hash": null,
                    "created_at": "2018-07-12T10:48:48.866Z"
                  }
                ]
              },
              "confirmed": {
                "GAS": "47320000000.0",
                "SWTH": "421549852102.0",
                "NEO": "50269113921.0"
              },
              "locked": {
                "GAS": "500000000.0",
                "NEO": "1564605000.0"
              }
            }
        """
        assert isinstance(addresses, list), "addresses should be a list object for eg. {}".format([addresses])
        assert isinstance(contract_hashes, list), "contract_hashes should be a list object for eg. [contract_hashes]"
        return balances._list_balances(self.base_url, addresses, contract_hashes)

    def list_trades(self, contract_hash, pair, from_time=None, to_time=None, limit=None):
        """Retrieve trades that have already occurred on Switcheo Exchange filtered by the request parameters.

        Args:
            contract_hash (str) : Only return trades for this contract hash.
            pair (str)          : Only return trades for this pair.
            from_time (int)     : Only return trades after this time in epoch seconds.
            to_time (int)       : Only return trades before this time in epoch seconds.
            limit (int)         : Only return this number of trades (min: 1, max: 10000, default: 5000).

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response:
            [
              {
                "id": "712a5019-3a23-463e-b0e1-80e9f0ad4f91",
                "fill_amount": 9122032316,
                "take_amount": 20921746,
                "event_time": "2018-06-08T11:32:03.219Z",
                "is_buy": false
              },
              {
                "id": "5d7e42a2-a8f3-40a9-bce5-7304921ff691",
                "fill_amount": 280477933,
                "take_amount": 4207169,
                "event_time": "2018-06-08T11:31:42.200Z",
                "is_buy": false
              },
              ...
            ]
        """
        return trades._list_trades(base_url=self.base_url, contract_hash=contract_hash, pair=pair,
                                   from_time=from_time, to_time=to_time, limit=limit)

    def deposit(self, priv_key_wif, asset_id, amount, contract_hash, blockchain="NEO"):
        """This api creates a deposit of provided asset on smart-contract.

        To be able to make a deposit, sufficient funds are required in the depositing wallet.
        This method performs two tasks 1. Creates a deposit on smart-contract 2. Executes it.

        Args:
            priv_key_wif (str)  : The private key wif of the user.
            asset_id (str)      : The asset symbol or ID to deposit. for eg. SWTH
            amount (int)        : Amount of tokens to deposit.
            contract_hash (str) : Switcheo Exchange contract hash to execute the deposit on.
            blockchain (str)    : Blockchain that the token to deposit is on. Possible values are: neo.

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object
        """
        deposits_resp = deposits._create_deposit(base_url=self.base_url, priv_key_wif=priv_key_wif,
                                                 asset_id=asset_id, amount=amount, contract_hash=contract_hash,
                                                 blockchain=blockchain)
        return deposits._execute_deposit(base_url=self.base_url,
                                         deposit=response_to_json(deposits_resp), priv_key_wif=priv_key_wif,)

    def list_offers(self, blockchain, pair, contract_hash):
        """Retrieves the best 70 offers (per side) on the offer book.

        Args:
            blockchain (str)   : Only return offers from this blockchain. Possible values are neo.
            pair (str)         : Only return offers from this pair, for eg. SWTH_NEO
            contract_hash (str): Only return offers for contract hash. e.g. eed0d2e14b0027f5f30ade45f2b23dc57dd54ad2

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response:
            [
                {
                 "id": "b3a91e19-3726-4d09-8488-7c22eca76fc0",
                 "offer_asset": "SWTH",
                 "want_asset": "NEO",
                 "available_amount": 2550000013,
                 "offer_amount": 4000000000,
                 "want_amount": 320000000
                }
            ]
        """
        return offers._list_offers(self.base_url, blockchain, pair, contract_hash)

    def list_orders(self, address, contract_hash, pair=None):
        """Retrieves the best 70 offers (per side) on the offer book.

        Args:
            address (str)      : Only returns orders made by this address.
            contract_hash (str): Only return offers for the contract hash. e.g. eed0d2e14b0027f5f30ad2b23dc57dd54ad2
            pair (str)         : The pair to buy or sell on. (optional)

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

            Example response:
            [
              {
                "id": "c415f943-bea8-4dbf-82e3-8460c559d8b7",
                "blockchain": "neo",
                "contract_hash": "c41d8b0c30252ce7e8b6d95e9ce13fdd68d2a5a8",
                "address": "20abeefe84e4059f6681bf96d5dcb5ddeffcc377",
                "side": "buy",
                "offer_asset_id": "c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b",
                "want_asset_id": "602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7",
                "offer_amount": "100000000",
                "want_amount": "20000000",
                "transfer_amount": "0",
                "priority_gas_amount": "0",
                "use_native_token": false,
                "native_fee_transfer_amount": 0,
                "deposit_txn": null,
                "created_at": "2018-05-15T10:54:20.054Z",
                "status": "processed",
                "fills": [...],
                "makes": [...]
              }
            ]
        """
        return orders._list_orders(self.base_url, address, contract_hash, pair=None)

    def create_order(self, priv_key_wif, pair, side, price, want_amount, asset_id,
                     use_native_tokens, contract_hash, blockchain="neo", order_type='limit'):
        """Create an order on SWTH DEX.

        Orders can only be created after sufficient funds have been deposited into the user's contract
        balance. A successful order will have zero or one make and/or zero or more fills.

        NOTE: Based on the params you are using, let's say you are trying to sell SWTH for NEO at the price of 0.01
        exchange rate. The `want_amount` for a sell would be the amount of NEO you want. For eg
        we want_amount of neo = 0.1 and we want to sell 1 SWTH for 0.0005 NEOs. In this case the sell order would
        become at 0.0005 exchange rate for 0.1 NEO and 200 SWTH ( so you need to have 200 SWTH in your smart-contract)

        Args:
            base_url (str)           : This paramter governs whether to connect to test or mainnet.
            priv_key_wif (str)       : The private key wif of the user.
            pair (str)               : The pair to buy or sell on.
            blockchain (str)         : Blockchain that the pair is on. Possible values are: neo.
            side (str)               : Whether to buy or sell on this pair. Possible values are: buy, sell.
            price (str)              : Buy or sell price to 8 decimal places precision.
            want_amount (int)        : Amount of tokens offered in the order.
            asset_id (str)           : Asset which is being traded for eg. in SWTH_NEO then its SWTH
            use_native_tokens (bool) : Whether to use SWTH as fees or not. Possible values are: true or false.
            order_type (str)         : Order type, possible values are: limit.
            contract_hash (str)      : Switcheo Exchange contract hash to execute the deposit on.

        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object
        """
        orders_response = orders._create_order(base_url=self.base_url, priv_key_wif=priv_key_wif, pair=pair,
                                               blockchain=blockchain, side=side, price=price, want_amount=want_amount,
                                               use_native_tokens=use_native_tokens, asset_id=asset_id,
                                               order_type=order_type, contract_hash=contract_hash,
                                               )

        orders_json_resp = response_to_json(orders_response)

        return orders._execute_order(base_url=self.base_url, order=orders_json_resp, priv_key_wif=priv_key_wif,)

    def withdraw(self, priv_key_wif, asset_id, amount, contract_hash, blockchain="NEO"):
        """Withdraw your balanaces from Switcheo smart contract balance.

        This function creates a withdrawal which is later executed.

        To be able to make a withdrawal, sufficient funds are required in the contract balance.
        A signature of the request payload has to be provided for this API call.

        Args:
            priv_key_wif (str)      : The private key wif of the user.
            asset_id (str)          : The asset symbol or ID to withdraw. for eg. SWTH
            amount (int)            : Amount of tokens to withdraw.
            contract_hash (str)     : Switcheo Exchange contract hash to execute the withdraw on.
            blockchain (str)        : Blockchain that the token to withdraw is on. Possible values are: neo.

        Returns:
            An id representing this transaction
            Example response:
            {
              "id": "e0f56e23-2e11-4848-b749-a147c872cbe6"
            }

        """
        withdrawals_response = withdrawals._create_withdrawal(base_url=self.base_url, asset_id=asset_id,
                                                              contract_hash=contract_hash, amount=amount,
                                                              priv_key_wif=priv_key_wif, blockchain=blockchain)
        withdrawals_response_json_obj = response_to_json(withdrawals_response)
        # Now lets execute withdrawal
        return withdrawals._execute_withdrawal(base_url=self.base_url,
                                               withdrawal=withdrawals_response_json_obj,
                                               priv_key_wif=priv_key_wif)

    def create_cancellation(self, order_id, priv_key_wif):
        """This API is responsible for order cancellation.

        Only orders with makes and with an available_amount of more than 0 can be cancelled.

        Args:
            order_id (str) : The order id which needs to be cancelled.
        Returns:
            If response from the server is HTTP_OK (200) then this returns the requests.response object

        """
        cancellation_response = orders._create_cancellation(self.base_url, order_id, priv_key_wif)
        cancellation_resp_json = response_to_json(cancellation_response)
        return orders._execute_cancellation(self.base_url, cancellation_resp_json, priv_key_wif)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Orders are instructions to buy or sell assets on Switcheo Exchange."""

# Retrieves orders from a specific address filtered by the given parameters.
LIST_ORDERS = "/orders"

# This endpoint creates an order which can be executed through Broadcast Order.
CREATE_ORDER = "/orders"

# This is the second endpoint required to execute an order. After using the Create Order endpoint,
# you will receive a response which needs to be signed.
EXECUTE_ORDER = "/orders/{id}/broadcast"

# This is the first API call required to cancel an order.
CREATE_CANCELLATION = "/cancellations"

# This is the second endpoint that must be called to cancel an order. After calling the Create Cancellation
# endpoint, you will receive a transaction in the response which must be signed.
EXECUTE_CANCELLATION = "/cancellations/{id}/broadcast"



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Withdrawals related URIs"""

# This endpoint creates a withdrawal which can be executed through Execute Withdrawal. To be able to make a
# withdrawal, sufficient funds are required in the contract balance.
CREATE_WITHDRAWAL = "/withdrawals"


# This is the second endpoint required to execute a withdrawal. After using the Create Withdrawal endpoint, you
# will receive a response which requires additional signing.
EXECUTE_WITHDRAWAL = "/withdrawals/{id}/broadcast"



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Implementation for custom datatypes to interact with the blockchain."""
from decimal import Decimal
from pyswitcheo.crypto_utils import reverse_hex

import logging

logger = logging.getLogger(__name__)


class Fixed8(object):
    """Fixed point representation of a given input number."""

    def __init__(self, value):
        """."""
        self.__value = float(Decimal(value).quantize(Decimal("1.00000000")))

    def to_hex(self):
        output = hex(round(self.__value * 1e+8))[2:]
        return "0" * (16 - len(output)) + output

    def to_reverse_hex(self):
        """Get a reverse hex representation of a given Fixed8."""
        return reverse_hex(self.to_hex())

    @property
    def value(self):
        """Return the underlying value of fixed 8."""
        return str(self.__value)

    @staticmethod
    def num_to_fixed_8(number, size=8):
        """Convert a given number to Fixed8 representation.

        Args:
            number (float/int) : Input which needs to be converted to a Fixed8 representation.
        Returns:
            Given number in Fixed8 representation.
        Raises:
            TypeError in case size param is not an integer

        """
        if size % 1 != 0:
            raise TypeError(
                "size param must be a whole integer. Received {size}".format(size=size)
            )
        return Fixed8(number).to_reverse_hex()[: size * 2]



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Crypto related wrapper functions."""

import re
import string
import logging
import base58
import hashlib
import binascii
from neocore.KeyPair import KeyPair

logger = logging.getLogger(__name__)

__HEX_STRING_REGEX_OBJ = re.compile(r"^([0-9A-Fa-f]{2})*$/")
MAX_TRANSACTION_ATTRIBUTE_SIZE = 65535


def is_hex(input_hex):
    """Check if the passed string is a hex string.

    Empty string is always treated as hex.
    """
    if input_hex == "":
        return True

    return all(c in set(string.hexdigits) for c in input_hex)


def ensure_hex(input_hex):
    """Check if the passed string is a hex string else raise exception.

    Empty string is always treated as hex.
    """

    if not is_hex(input_hex):
        raise Exception(
            "Expected a hexstring but got {input_hex}.".format(input_hex=input_hex)
        )


def reverse_hex(input_hex):
    """Reverse a HEX string, treating 2 chars as a byte.

    Expected results look like this:
    reverse_hex('abcdef') = 'efcdab'

    Args:
        input_hex (str): Input string in hex which needs to be converted.

    Returns:
        Hex string reversed.
    """
    return "".join([input_hex[x: x + 2] for x in range(0, len(input_hex), 2)][::-1])


def num_to_hex_string(num, size=1, little_endian=False):
    """Convert a given number to hex string.

    Converts a number to a big endian hexstring of a suitable size, optionally little endian

    Args:
        num (int)   : Input int for which we need to get the hex string
        size  (int) : The required size in bytes, eg 1 for Uint8, 2 for Uint16. Defaults to 1.
    Returns:
        (str)

    """
    if num < 0:
        raise Exception("num should be unsigned (>= 0)")

    if size % 1 != 0:
        raise TypeError("size must be a whole integer")

    size = size * 2
    hexstring = hex(num)[2:]
    hexstr_len = len(hexstring)

    output = (
        hexstring if hexstr_len % size == 0 else ("0" * (size) + hexstring)[hexstr_len:]
    )
    if little_endian:
        return reverse_hex(output)
    return output


def num_to_var_int(num):
    """Convert a number to a variable length Int. Used for array length header.

    Detailed explanation

    Args:
        num (int) : A number
    Returns:
        (str) hexstring of the variable Int
    """
    if num < 0xfd:
        return num_to_hex_string(num)
    elif num <= 0xffff:
        # uint16
        return "fd" + num_to_hex_string(num, 2, True)
    elif num <= 0xffffffff:
        # uint32
        return "fe" + num_to_hex_string(num, 3, True)
    else:
        # uint64
        return "ff" + num_to_hex_string(num, 8, True)


def encode_msg(msg):
    """Convert a given msg to its hex representation.

    This is generally used when we send signed payload to the Switcheo api.

    Args:
        msg (str) : Input message which needs to be encoded.
    Returns:
        encoded message (str)
    """
    if isinstance(msg, str):
        msg = msg.encode("UTF-8")

    encoded_msg = binascii.hexlify(msg)
    length_hex = hex(int(len(encoded_msg) / 2))[2:]
    logger.debug("Length of hex message {0}".format(length_hex))

    encoded_msg = "010001f0{hex_len}{enc_msg}0000".format(
        hex_len=length_hex, enc_msg=encoded_msg.decode()
    )
    logger.debug("Final message to sign : {0}".format(encoded_msg))
    return encoded_msg


def get_private_key_from_wif(wif):
    """Fetch the private key from a wif represented in string format.

    Args:
        wif (str) : wif from which we need to extract the private key

    Returns:
        private key in bytearray format
    """
    pk = KeyPair.PrivateKeyFromWIF(wif)
    return KeyPair(pk).PrivateKey


def get_wif_from_private_key(priv_key):
    """Convert the given privatekey to a wif format.

    Args:
        priv_key (str) : private key in its hex string format.
    Returns:
        WIF format
    """

    # Step 1: let's add 80 in front of it and 01 in the end
    extended_key = "80" + priv_key + "01"
    # Step 2: first SHA-256
    first_sha256 = hashlib.sha256(binascii.unhexlify(extended_key)).hexdigest()
    # Step 3: second SHA-256
    second_sha256 = hashlib.sha256(binascii.unhexlify(first_sha256)).hexdigest()
    # Step 4-5: add checksum to end of extended key
    final_key = extended_key + second_sha256[:8]
    # Step 6: finally the Wallet Import Format is the base 58 encode of final_key
    return base58.b58encode(binascii.unhexlify(final_key))


def get_script_hash_from_wif(wif):
    """Fetch the script hash of the public key from a wif represented in string format.

    Args:
        wif (str) : wif from which we need to extract the public key script hash

    Returns:
        public key script hash in string format
    """
    pk = KeyPair.PrivateKeyFromWIF(wif)
    keypair = KeyPair(pk)
    logger.debug("Public Address is {}".format(keypair.GetAddress()))
    return get_script_hash_from_address(keypair.GetAddress())


def get_script_hash_from_address(address):
    """Convert a given address to script hash.
    This code has been taken from:
    https://github.com/CityOfZion/neo-python-core/blob/fcb0837e8f69e6f4dc01f2861b856affd2213446/neocore/bin/cli.py#L23
    """
    data = bytes(base58.b58decode(address))

    # Make sure signature byte is correct. In python 3, data[0] is bytes, and in 2 it's str.
    # We use this isinstance checke to make it work with both Python 2 and 3.
    is_correct_signature = data[0] != 0x17 if isinstance(data[0], bytes) else b"\x17"
    if not is_correct_signature:
        raise Exception("Invalid address: wrong signature byte")

    # Make sure the checksum is correct
    if data[-4:] != hashlib.sha256(hashlib.sha256(data[:-4]).digest()).digest()[:4]:
        raise Exception("Invalid address: invalid checksum")

    # Return only the scripthash bytes, reverse it and return hex of it.
    return data[1:-4][::-1].hex()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Deposits allows movement of funds from your wallet into the Switcheo smart contract.
This movement of funds is free of charge. Trading on Switcheo Exchange can only be done
using funds that have been successfully deposited into the smart contract.
Deposits are not instantaneous. Once a deposit has been executed, funds in your wallet
balance would be deducted by the amount of tokens you chose to deposit. Funds deducted
from your wallet balance will be added to your contract balance but put on hold until
Switcheo has determined that it has been successfully broadcasted to the blockchain.
"""
import logging
import requests
from pyswitcheo import utils
from pyswitcheo.internal.urls import deposits
from pyswitcheo.serialization import sign_msg, serialize_transaction
from pyswitcheo.crypto_utils import (
    encode_msg,
    get_private_key_from_wif,
    get_script_hash_from_wif,
)

logger = logging.getLogger(__name__)

# TODO: (ansrivas) Give an option to load from wallet, pass wif and pass private key


def _create_deposit(base_url, priv_key_wif, asset_id, amount, contract_hash, blockchain="NEO"):
    """This endpoint creates a deposit which can be executed through Execute Deposit.

    To be able to make a deposit, sufficient funds are required in the depositing wallet.
    NOTE: After calling this endpoint, the Execute Deposit endpoint has to be called for the deposit to be executed.

    Args:
        base_url (str)      : This paramter governs whether to connect to test or mainnet..
        priv_key_wif (str)  : The private key wif of the user.
        asset_id (str)      : The asset symbol or ID to deposit. for eg. SWTH
        amount (int)        : Amount of tokens to deposit.
        contract_hash (str) : Switcheo Exchange contract hash to execute the deposit on.
        blockchain (str)    : Blockchain that the token to deposit is on. Possible values are: neo.

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object
        Check schemas.CREATE_DEPOSIT_SCHEMA
    """

    signable_params = {
        "blockchain": blockchain,
        "asset_id": asset_id,
        "amount": utils.convert_to_neo_asset_amount(amount, asset_id, base_url),
        "timestamp": utils.get_current_epoch_milli(),
        "contract_hash": contract_hash,
    }
    signable_params_json_str = utils.jsonify(signable_params)

    # The depositer's address. Do not include this in the parameters to be signed.
    # This needs to be the script hash of the public key
    script_hash = get_script_hash_from_wif(priv_key_wif)

    pk = get_private_key_from_wif(priv_key_wif)
    encoded_msg = encode_msg(signable_params_json_str)
    signature = sign_msg(encoded_msg, pk)

    params = {**signable_params, "signature": signature, "address": script_hash}

    logger.debug("Params being sent to create deposit: {0}".format(params))
    url = utils.format_urls(base_url, deposits.CREATE_DEPOSIT)
    resp = requests.post(url, json=params)
    return utils.response_else_exception(resp)


def _execute_deposit(base_url, deposit, priv_key_wif):
    """This is the second endpoint required to execute a deposit. After using the Create Deposit endpoint,
    you will receive a response which requires additional signing.
    The signature should then be attached as the signature parameter in the request payload.

    Note that a sha256 parameter is provided for convenience to be used directly as part of the ECDSA signature
    process. In production mode, this should be recalculated for additional security.

    Args:
        base_url (str)     : This paramter governs whether to connect to test or mainnet..
        deposit (json)     : The correct json response returned from create_deposit function (json.loads)
        priv_key_wif (str) : The private key WIF of the user.

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

    """
    logger.debug("Deposit transaction is {0}".format(utils.jsonify(deposit["transaction"])))

    pk = get_private_key_from_wif(priv_key_wif)
    # signature is Signed response from create deposit endpoint.
    # signable_params_json_str = utils.jsonify(deposit["transaction"])

    signature = sign_msg(serialize_transaction(deposit["transaction"], False), pk)

    url = utils.format_urls(base_url, deposits.EXECUTE_DEPOSIT.format(id=deposit["id"]))
    resp = requests.post(url, json={"signature": signature})
    return utils.response_else_exception(resp)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Orders are instructions to buy or sell assets on Switcheo Exchange.
At the moment, only Limit orders are available. Market, Fill-Or-Cancel, Make-Or-Cancel, etc. strategies are not
available yet.
As such, orders will contain a combination of zero or one make and/or zero or more fills.
Once an order is placed, the funds required for the order is removed from the user's balance and placed on hold
until the order is filled or the order is cancelled.
"""

import logging
import requests
from pyswitcheo import utils
from pyswitcheo.internal.urls import orders
from pyswitcheo.datatypes.fixed8 import Fixed8
from pyswitcheo.serialization import sign_msg, sign_array, sign_transaction
from pyswitcheo.crypto_utils import (
    encode_msg,
    get_script_hash_from_address,
    get_private_key_from_wif,
    get_script_hash_from_wif,
)

logger = logging.getLogger(__name__)


def _list_orders(base_url, address, contract_hash, pair=None):
    """Retrieves the best 70 offers (per side) on the offer book.

    Args:
        base_url (str)     : This paramter governs whether to connect to test or mainnet.
        address (str)      : Only returns orders made by this address.
        contract_hash (str): Only return offers for the contract hash. e.g. eed0d2e14b0023dc57dd54ad2
        pair (str)         : The pair to buy or sell on. (optional)

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response:
        [
          {
            "id": "c415f943-bea8-4dbf-82e3-8460c559d8b7",
            "blockchain": "neo",
            "contract_hash": "c41d8b0c30252ce7e8b6d95e9ce13fdd68d2a5a8",
            "address": "20abeefe84e4059f6681bf96d5dcb5ddeffcc377",
            "side": "buy",
            "offer_asset_id": "c56f33fc6ecfcd0c225c4ab356fee0faebe74a6daff7c9b",
            "want_asset_id": "602c79718b16e442de58778e148d0bb16cee7969282de7",
            "offer_amount": "100000000",
            "want_amount": "20000000",
            "transfer_amount": "0",
            "priority_gas_amount": "0",
            "use_native_token": false,
            "native_fee_transfer_amount": 0,
            "deposit_txn": null,
            "created_at": "2018-05-15T10:54:20.054Z",
            "status": "processed",
            "fills": [...],
            "makes": [...]
          }
        ]
    """
    address = get_script_hash_from_address(address)
    params = {"address": address, "contract_hash": contract_hash}

    if pair:
        params["pair"] = pair
    url = utils.format_urls(base_url, orders.LIST_ORDERS)
    resp = requests.get(url, params=params)
    return utils.response_else_exception(resp)


def _create_order(
    base_url,
    priv_key_wif,
    pair,
    blockchain,
    side,
    price,
    want_amount,
    asset_id,
    use_native_tokens,
    order_type,
    contract_hash,
):
    """This endpoint creates an order which can be executed through Broadcast Order.

    Orders can only be created after sufficient funds have been deposited into the user's contract
    balance. A successful order will have zero or one make and/or zero or more fills.

    NOTE: Based on the params you are using, let's say you are trying to sell SWTH for NEO at the price of 0.01
    exchange rate. The `want_amount` for a sell would be the amount of NEO you want. For eg
    we want_amount of neo = 0.1 and we want to sell 1 SWTH for 0.0005 NEOs. In this case the sell order would become
    at 0.0005 exchange rate for 0.1 NEO and 200 SWTH ( so you need to have 200 SWTH in your smart-contract)

    IMPORTANT: After calling this endpoint, the Broadcast Order endpoint has to be called for the order to be executed.

    Args:
        base_url (str)           : This paramter governs whether to connect to test or mainnet.
        priv_key_wif (str)       : The private key wif of the user.
        pair (str)               : The pair to buy or sell on.
        blockchain (str)         : Blockchain that the pair is on. Possible values are: neo.
        address (str)            : Address of the order maker. Do not include this in the parameters to be signed.
        side (str)               : Whether to buy or sell on this pair. Possible values are: buy, sell.
        price (str)              : Buy or sell price to 8 decimal places precision.
        want_amount (int)        : Amount of tokens offered in the order.
        asset_id (str)           : Asset which is being traded for eg. in SWTH_NEO then its SWTH
        use_native_tokens (bool) : Whether to use SWTH as fees or not. Possible values are: true or false.
        order_type (str)         : Order type, possible values are: limit.
        contract_hash (str)      : Switcheo Exchange contract hash to execute the deposit on.

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object
        Check schemas.CREATE_ORDER_RESPONSE_SCHEMA
    """
    # The current timestamp to be used as a nonce as epoch milliseconds.
    timestamp = utils.get_current_epoch_milli()

    # TODO: (ansrivas) Check how to handle currencies which are not divisible, for eg. NEO ( until v3 is released)
    want_amount = utils.convert_to_neo_asset_amount(want_amount, asset_id, base_url)
    price = Fixed8(price).value
    signable_params = {
        "blockchain": blockchain,
        "price": price,
        "side": side,
        "want_amount": want_amount,
        "use_native_tokens": use_native_tokens,
        "timestamp": timestamp,
        "order_type": order_type,
        "contract_hash": contract_hash,
        "pair": pair,
    }

    # The order creator's address. Do not include this in the parameters to be signed.
    # This needs to be the script hash of the public key
    script_hash = get_script_hash_from_wif(priv_key_wif)
    pk = get_private_key_from_wif(priv_key_wif)

    signable_params_json_str = utils.jsonify(signable_params)
    encoded_msg = encode_msg(signable_params_json_str)
    signature = sign_msg(encoded_msg, pk)

    params = {**signable_params, "signature": signature, "address": script_hash}

    logger.debug("Params being sent to create orders: {0}".format(params))
    url = utils.format_urls(base_url, orders.CREATE_ORDER)
    resp = requests.post(url, json=params)
    return utils.response_else_exception(resp)


def _execute_order(base_url, order, priv_key_wif):
    """This is the second endpoint required to execute an order.

    After using the Create Order endpoint, you will receive a response which needs to be signed.
    Every txn of the fills and makes in the Create Order response should be signed.

    Args:
        order (dict) : The response object returned after creating an order
    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

    """
    id = order["id"]
    fills, makes = order["fills"], order["makes"]
    logger.debug("Fills are {}".format(fills))
    logger.debug("makes are {}".format(makes))

    priv_key = get_private_key_from_wif(priv_key_wif)
    signatures = {
        "fills": sign_array(fills, priv_key),
        "makes": sign_array(makes, priv_key),
    }
    params = {"signatures": signatures}

    logger.debug("Params being sent to create deposit: {0}".format(params))
    url = utils.format_urls(base_url, orders.EXECUTE_ORDER.format(id=id))

    resp = requests.post(url, json=params)
    return utils.response_else_exception(resp)


def _create_cancellation(base_url, order_id, priv_key_wif):
    """This is the first API call required to cancel an order.

    Only orders with makes and with an available_amount of more than 0 can be cancelled.

    Args:
        order_id (str) : The order id which needs to be cancelled.
    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

    """
    signable_params = {
        "order_id": order_id,
        "timestamp": utils.get_current_epoch_milli(),
    }

    priv_key = get_private_key_from_wif(priv_key_wif)

    signable_params_json_str = utils.jsonify(signable_params)
    encoded_msg = encode_msg(signable_params_json_str)
    signature = sign_msg(encoded_msg, priv_key)

    script_hash = get_script_hash_from_wif(priv_key_wif)
    params = {**signable_params, "signature": signature, "address": script_hash}

    logger.debug("Params being sent to create deposit: {0}".format(params))
    url = utils.format_urls(base_url, orders.CREATE_CANCELLATION)
    resp = requests.post(url, json=params)
    return utils.response_else_exception(resp)


def _execute_cancellation(base_url, cancellation, priv_key_wif):
    """This is the second endpoint that must be called to cancel an order.

    After calling the Create Cancellation endpoint, you will receive a transaction in the resp which must be signed.

    Args:
        cancellation (dict) : The cancellation object which is received from create cancellation end point.
    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

    """
    logger.debug("cancellation transaction is {0}".format(utils.jsonify(cancellation["transaction"])))

    pk = get_private_key_from_wif(priv_key_wif)
    signature = sign_transaction(cancellation["transaction"], pk)

    url = utils.format_urls(base_url, orders.EXECUTE_CANCELLATION.format(id=cancellation["id"]))
    resp = requests.post(url, json={"signature": signature})
    return utils.response_else_exception(resp)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Implementation for custom datatypes to interact with the blockchain."""

from collections import namedtuple

# Represents a TransactionAttribute
TransactionAttribute = namedtuple("TransactionAttribute", ["usage", "data"])

# Represents a TransactionInput
TransactionInput = namedtuple("TransactionInput", ["prevHash", "prevIndex"])

# Represents a TransactionOutput
TransactionOutput = namedtuple("TransactionOutput", ["assetId", "value", "scriptHash"])

# Represents a Witness
Witness = namedtuple("Witness", ["invocationScript", "verificationScript"])



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""An offer represents an open order that rests on the Switcheo Exchange offer book.
Funds used to make an offer will be placed on hold unless the order is cancelled or filled
"""
import requests
from pyswitcheo import utils
from pyswitcheo.internal.urls.offers import LIST_OFFERS


def _list_offers(base_url, blockchain, pair, contract_hash):
    """Retrieves the best 70 offers (per side) on the offer book.

    Args:
        base_url (str): This paramter governs whether to connect to test or mainnet..
        blockchain (str)   : Only return offers from this blockchain. Possible values are neo.
        pair (str)         : Only return offers from this pair, for eg. SWTH_NEO
        contract_hash (str): Only return offers for the contract hash. e.g. eed0d2e14b0027f5f30ade45f2b23dc57dd54ad2

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response:
        [
            {
             "id": "b3a91e19-3726-4d09-8488-7c22eca76fc0",
             "offer_asset": "SWTH",
             "want_asset": "NEO",
             "available_amount": 2550000013,
             "offer_amount": 4000000000,
             "want_amount": 320000000
            }
        ]
    """
    params = {"blockchain": blockchain, "pair": pair, "contract_hash": contract_hash}
    url = utils.format_urls(base_url, LIST_OFFERS)
    resp = requests.get(url, params=params)
    return utils.response_else_exception(resp)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""This section consists of endpoints that allow retrieval of Switcheo Exchange information.
Authentication is not required for these endpoints.
"""

import logging
import requests
from pyswitcheo import utils
from pyswitcheo.internal.urls import exchange

logger = logging.getLogger(__name__)


def _list_currency_pairs(base_url, bases=None):
    """Fetch available currency pairs on Switcheo Exchange filtered by the base parameter. Defaults to all pairs.

    Args:
        base_url (str)    : This paramter governs whether to connect to test or mainnet..
        bases (list[str]) : Provides pairs for these base symbols. Possible values are NEO, GAS, SWTH, USD.

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response
        [
          "GAS_NEO",
          "SWTH_NEO"
        ]
    """
    params = {}
    if bases:
        params["bases"] = bases

    url = utils.format_urls(base_url, exchange.LIST_CURRENCY_PAIRS)
    resp = requests.get(url, params=params)
    return utils.response_else_exception(resp)


def _get_exchange_timestamp(base_url):
    """Returns the current timestamp in the exchange.

    This value should be fetched and used when a timestamp parameter is required for API requests.
    If the timestamp used for your API request is not within an acceptable range of the exchange's
    timestamp then an invalid signature error will be returned. The acceptable range might vary, but it
    should be less than one minute.

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response
        {
          "timestamp": 1534392760908
        }
    """
    url = utils.format_urls(base_url, exchange.GET_TIMESTAMP)
    resp = requests.get(url)
    return utils.response_else_exception(resp)


def _list_contracts(base_url):
    """Fetch updated hashes of contracts deployed by Switcheo.

    Args:
        base_url (str): This paramter governs whether to connect to test or mainnet.

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response
        {
          "NEO":
          {
            "V1": "0ec5712e0f7c63e4b0fea31029a28cea5e9d551f",
            "V1_5": "c41d8b0c30252ce7e8b6d95e9ce13fdd68d2a5a8",
            "V2": "48756743d524af03aa75729e911651ffd3cbe7d8"
          }
        }
    """
    url = utils.format_urls(base_url, exchange.LIST_CONTRACTS)
    resp = requests.get(url)
    return utils.response_else_exception(resp)


def _get_contract_tokens_info(base_url):
    """Fetch updated hashes of contracts deployed by Switcheo along with their precision.

    Args:
        base_url (str): This paramter governs whether to connect to test or mainnet.

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response
        {
          "NEO": {
            "hash": "c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b",
            "decimals": 8
          },
          "GAS": {
            "hash": "602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7",
            "decimals": 8
          },
          "SWTH": {
            "hash": "ab38352559b8b203bde5fddfa0b07d8b2525e132",
            "decimals": 8
          },
          ...
        }
    """
    url = utils.format_urls(base_url, exchange.GET_TOKENS)
    resp = requests.get(url)
    return utils.response_else_exception(resp)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""List contract balances of the given address and contract.
The purpose of this endpoint is to allow convenient querying of a user's balance across multiple blockchains,
for example, if you want to retrieve a user's NEO and ethereum balances.
As such, when using this endpoint, balances for the specified addresses and contract hashes will be merged and summed.
"""
import requests
from pyswitcheo import utils
from pyswitcheo.internal.urls.balances import LIST_BALANCES
from pyswitcheo.crypto_utils import get_script_hash_from_address


def _list_balances(base_url, addresses, contract_hashes):
    """List contract balances of the given address and contract hashes.

    NOTE: The address gets converted to scripthash internally
    Args:
        base_url (str)               : This paramter governs whether to connect to test or mainnet..
        addresses (list[str])        : Only return balances for these addresses.
        contract_hashes (list[str])  : Only return balances from these contract hashes.

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response:
        {
          "confirming": {
            "GAS": [
              {
                "event_type": "withdrawal",
                "asset_id": "602c79718b16e442de58778de6b7b16cee7969282de7",
                "amount": -100000000,
                "transaction_hash": null,
                "created_at": "2018-07-12T10:48:48.866Z"
              }
            ]
          },
          "confirmed": {
            "GAS": "47320000000.0",
            "SWTH": "421549852102.0",
            "NEO": "50269113921.0"
          },
          "locked": {
            "GAS": "500000000.0",
            "NEO": "1564605000.0"
          }
        }
    """
    addresses = [get_script_hash_from_address(address) for address in addresses]
    params = {"addresses": addresses, "contract_hashes": contract_hashes}
    url = utils.format_urls(base_url, LIST_BALANCES)
    resp = requests.get(url, params=params)
    return utils.response_else_exception(resp)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""The tickers section consists of endpoints that allow retrieval of market data on Switcheo Exchange.
Authentication is not required for these endpoints.
"""

# Returns candlestick chart data filtered by url parameters.
CANDLE_STICKS = "/tickers/candlesticks"

# Returns 24-hour data for all pairs and markets.
LAST_TWENTY_FOUR_HOURS = "/tickers/last_24_hours"

# Returns last price of given symbol(s). Defaults to all symbols.
LAST_PRICE = "/tickers/last_price"



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Custom exception raised by pyswitcheo"""

# Human readable http error codes
import json
from http import HTTPStatus

_ERR = {
    HTTPStatus.BAD_REQUEST: "Your request is badly formed.",
    HTTPStatus.UNAUTHORIZED: "You did not provide a valid signature.",
    HTTPStatus.NOT_FOUND: "The specified endpoint or resource could not be found.",
    HTTPStatus.NOT_ACCEPTABLE: "You requested a format that isn't json.",
    HTTPStatus.TOO_MANY_REQUESTS: "Slow down requests and use Exponential backoff timing.",
    HTTPStatus.UNPROCESSABLE_ENTITY: "Your request had validation errors.",
    HTTPStatus.INTERNAL_SERVER_ERROR: "We had a problem with our server. Try again later.",
    HTTPStatus.SERVICE_UNAVAILABLE: "We're temporarily offline for maintenance. Please try again later.",
}


class HTTPResponseError(Exception):
    """Wrapper around Exception to raise custom messages."""

    def __init__(self, response):
        self.value = {
            "code": response.status_code,
            "server_msg": response.text,
            "err_msg": _ERR.get(response.status_code, "Unexpected error"),
        }

    def __str__(self):
        return json.dumps(self.value, indent=4, sort_keys=True)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""The tickers section consists of endpoints that allow retrieval of market data on Switcheo Exchange.
Authentication is not required for these endpoints.
"""
import logging
import requests
from pyswitcheo import utils
from pyswitcheo.internal.urls import tickers

logger = logging.getLogger(__name__)


def _get_candle_sticks(base_url, pair, start_time, end_time, interval):
    """Get candlestick chart data filtered by url parameters.

    Args:
        base_url (str)     : This paramter governs whether to connect to test or mainnet.
        pair (str)	       : Show chart data of this trading pair
        start_time (int)   : Start of time range for data in epoch seconds
        end_time (int)	   : End of time range for data in epoch seconds
        interval (int)	   : Candlestick period in minutes Possible values are: 1, 5, 30, 60, 360, 1440


    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response:
        [
          {
            "time": "1531215240",
            "open": "0.00049408",
            "close": "0.00049238",
            "high": "0.000497",
            "low": "0.00048919",
            "volume": "110169445.0",
            "quote_volume": "222900002152.0"
          },
          {
            "time": "1531219800",
            "open": "0.00050366",
            "close": "0.00049408",
            "high": "0.00050366",
            "low": "0.00049408",
            "volume": "102398958.0",
            "quote_volume": "205800003323.0"
          },
          ...
        ]
    """
    url = utils.format_urls(base_url, tickers.CANDLE_STICKS)
    params = {
        "pair": pair,
        "start_time": start_time,
        "end_time": end_time,
        "interval": interval,
    }
    resp = requests.get(url, params=params)
    return utils.response_else_exception(resp)


def __get_prices(url, symbols=None, bases=None):
    """."""
    params = {}
    if symbols:
        params["symbols"] = symbols

    if bases:
        params["bases"] = bases

    logger.debug("Sending params in this request {params}".format(params=params))
    resp = requests.get(url, params=params)
    return utils.response_else_exception(resp)


def _get_last_twenty_four_hours(base_url):
    """Get 24-hour data for all pairs and markets.

    Args:
        base_url (str)  : This paramter governs whether to connect to test or mainnet.
    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

    """
    url = utils.format_urls(base_url, tickers.LAST_TWENTY_FOUR_HOURS)
    return __get_prices(url)


def _get_last_price(base_url, symbols=None):
    """Get last price of given symbol(s). Defaults to all symbols.

    Args:
        base_url (str)       : This paramter governs whether to connect to test or mainnet..
        symbols (list[str])  : Return the price for these symbols for eg. ["SWTH"]

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

    """
    url = utils.format_urls(base_url, tickers.LAST_PRICE)
    return __get_prices(url, symbols)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""A trade represents a fill of an offer.
This happens when an incoming order matches an offer on the opposite side of the order book in price.
Trades can be seen on the Trade History column on Switcheo Exchange.
"""

# Retrieves trades that have already occurred on Switcheo Exchange filtered by the request parameters.
LIST_TRADES = "/trades"



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""This section consists of endpoints that allow retrieval of Switcheo Exchange information.
Authentication is not required for these endpoints.
"""

# Returns the current timestamp in the exchange, this value should be
# fetched and used when a timestamp parameter is required for API requests.
GET_TIMESTAMP = "/exchange/timestamp"

# Returns updated hashes of contracts deployed by Switcheo.
LIST_CONTRACTS = "/exchange/contracts"

# Returns updated hashes of contracts deployed by Switcheo.
GET_TOKENS = "/exchange/tokens"

# Retrieve available currency pairs on Switcheo Exchange filtered by the base parameter.
# Defaults to all pairs. The valid base currencies are currently: NEO, GAS, SWTH.
LIST_CURRENCY_PAIRS = "/exchange/pairs"



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""An offer represents an open order that rests on the Switcheo Exchange offer book.
Funds used to make an offer will be placed on hold unless the order is cancelled or filled.
"""

# Retrieves the best 70 offers (per side) on the offer book.
LIST_OFFERS = "/offers"



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Crypto related wrapper functions."""

import logging
import pyswitcheo.crypto_utils as cutils
from neocore.Cryptography.Crypto import Crypto
from pyswitcheo.datatypes.fixed8 import Fixed8
from pyswitcheo.datatypes.transaction_types import (
    TransactionInput,
    TransactionOutput,
    TransactionAttribute,
    Witness,
)

logger = logging.getLogger(__name__)

MAX_TRANSACTION_ATTRIBUTE_SIZE = 65535


def serialize_transaction_output(output):
    """Serialize an object of type TransactionOutput

    TransactionOutput has three params
    1. assetId: assetId, Uint256
    2. value: value of output, Fixed8
    3. scriptHash of type Uint160

    Args:
        input (TransactionOutput) : TransactionOutput which needs to be serialized.
    Returns:
        (str) serialized version of TransactionOutput
    """
    value = Fixed8(output.value).to_reverse_hex()
    return "".join(
        [
            cutils.reverse_hex(output.assetId),
            value,
            cutils.reverse_hex(output.scriptHash),
        ]
    )


def serialize_transaction_input(input):
    """Serialize an object of type TransactionInput

    TransactionInput has two params
    1. prevHash: Transaction hash (Uint256)
    2. prevIndex:  Index of the coin in the previous transaction (Uint16)

    Args:
        input (TransactionInput) : TransactionInput which needs to be serialized.
    Returns:
        (str) serialized version of TransactionInput
    """
    logger.debug("serialize_transaction_input")
    logger.debug(cutils.reverse_hex(input.prevHash))
    logger.debug(cutils.reverse_hex(cutils.num_to_hex_string(input.prevIndex, 2)))
    return cutils.reverse_hex(input.prevHash) + cutils.reverse_hex(
        cutils.num_to_hex_string(input.prevIndex, 2)
    )


def serialize_claim_exclusive():
    """."""
    pass


def serialize_contract_exclusive():
    """."""
    pass


def serialize_invocation_exclusive(transaction):
    """Short explanation.

    Detailed explanation

    Args:

    Returns:

    """
    if transaction["type"] != 0xd1:
        raise TypeError("TransactionInvocation type should be 0xd1.")
    out = cutils.num_to_var_int(int(len(transaction["script"]) / 2))
    out += transaction["script"]
    if transaction["version"] >= 1:
        out += str(Fixed8.num_to_fixed_8(transaction["gas"]))
    return out


def serialize_exclusive(transaction_type):
    """."""
    op_type = {
        2: serialize_claim_exclusive,
        128: serialize_contract_exclusive,
        209: serialize_invocation_exclusive,
    }
    return op_type[transaction_type]


def serialize_witness(witness):
    """Serialize an object of type Witness

    Witness object has two params
    1. invocationScript:  This data is stored as is (Little Endian)
    2. verificationScript: This data is stored as is (Little Endian)

    Args:
        input (witness) : witness which needs to be serialized.
    Returns:
        (str) serialized version of witness
    """
    invo_len = cutils.num_to_var_int(len(witness.invocationScript) / 2)
    veri_len = cutils.num_to_var_int(len(witness.verificationScript) / 2)
    return invo_len + witness.invocationScript + veri_len + witness.verificationScript


def sign_transaction(transaction, priv_key):
    """Sign a transaction object returned as a part of any transaction creation using user's private key.

    Args:
        transaction (json) : Transaction is dictionary object which is returned after creating a transaction.
        priv_key (bytes)  : Private key to be used to sign this.
    Returns:
        A signed transaction string
    """
    serialized_tx_msg = serialize_transaction(tx=transaction, signed=False)
    return sign_msg(serialized_tx_msg, priv_key)


def sign_msg(msg, priv_key):
    """Sign a given message using a private key.

    Args:
        msg (str)         : Message which needs to be signed.
        priv_key (bytes)  : Private key to be used to sign this.

    Returns:
        Signed message as a byte array.
    """
    return Crypto.Sign(msg, priv_key).hex()


def serialize_transaction_attribute(attr):
    """Serialize a TransactionAttribute

    Detailed explanation

    Args:

    Returns:
        str
    """
    attr_len = len(attr.data)
    if attr_len > MAX_TRANSACTION_ATTRIBUTE_SIZE:
        raise Exception(
            "Attribute data size is beyond max attribute size {0}".format(
                MAX_TRANSACTION_ATTRIBUTE_SIZE
            )
        )

    out = cutils.num_to_hex_string(attr.usage)
    if attr.usage == 0x81:
        out += cutils.num_to_hex_string(attr_len / 2)
    elif attr.usage == 0x90 or attr.usage >= 0xf0:
        out += cutils.num_to_var_int(attr_len / 2)

    if (attr.usage == 0x02) or (attr.usage == 0x03):
        out += attr.data[2:64]
    else:
        out += attr.data

    return out


def serialize_transaction(tx, signed=True):
    """Serialize a transaction object

    Whenever an operation is invoked on the blockchain, we get a transaction object.
    As a rest response we can pass this here to sign it.

    Args:

    Returns:

    """
    tx_out = tx["outputs"]
    tx_ins = tx["inputs"]
    tx_scripts = tx["scripts"]

    out = ""
    out += cutils.num_to_hex_string(tx["type"])

    out += cutils.num_to_hex_string(tx["version"])

    out += (serialize_exclusive(tx["type"]))(tx)

    out += cutils.num_to_var_int(len(tx["attributes"]))

    for attribute in tx["attributes"]:
        attr = TransactionAttribute(**attribute)
        out += serialize_transaction_attribute(attr)

    out += cutils.num_to_var_int(len(tx_ins))

    for tx_in in tx_ins:
        inp = TransactionInput(**tx_in)
        out += serialize_transaction_input(inp)

    out += cutils.num_to_var_int(len(tx_out))

    for output in tx_out:
        outp = TransactionOutput(**output)
        out += serialize_transaction_output(outp)

    if signed and tx_scripts and (len(tx_scripts) > 0):
        out += cutils.num_to_var_int(len(tx_scripts))
        for script in tx_scripts:
            witness = Witness(**script)
            out += serialize_witness(witness)

    logger.debug("Final serialized transaction message to sign {0}".format(out))
    return out.strip()


def sign_array(input_arr, priv_key):
    """Sign each item in an input array.

    Args:
        input_arr (dict) : An input array with transaction objects. This is a dictionary with "txn" key in it.
    Returns:
        A dictionary of signed objects, where key is the id of each element in the input_arr.
    """
    signed_map = {}
    for item in input_arr:
        signed_map[item["id"]] = sign_transaction(item["txn"], priv_key)
    return signed_map



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""A trade represents a fill of an offer.
This happens when an incoming order matches an offer on the opposite side of the order book in price.
Trades can be seen on the Trade History column on Switcheo Exchange.
"""
import logging
import requests
from pyswitcheo import utils
from pyswitcheo.internal.urls import trades

logger = logging.getLogger(__name__)


def _list_trades(
    base_url, contract_hash, pair, from_time=None, to_time=None, limit=None
):
    """Retrieve trades that have already occurred on Switcheo Exchange filtered by the request parameters.

    Args:
        base_url (str): This paramter governs whether to connect to test or mainnet.
        contract_hash (str) : Only return trades for this contract hash.
        pair (str)          : Only return trades for this pair.
        from_time (int)     : Only return trades after this time in epoch seconds.
        to_time (int)       : Only return trades before this time in epoch seconds.
        limit (int)         : Only return this number of trades (min: 1, max: 10000, default: 5000).

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

        Example response:
        [
          {
            "id": "712a5019-3a23-463e-b0e1-80e9f0ad4f91",
            "fill_amount": 9122032316,
            "take_amount": 20921746,
            "event_time": "2018-06-08T11:32:03.219Z",
            "is_buy": false
          },
          {
            "id": "5d7e42a2-a8f3-40a9-bce5-7304921ff691",
            "fill_amount": 280477933,
            "take_amount": 4207169,
            "event_time": "2018-06-08T11:31:42.200Z",
            "is_buy": false
          },
          ...
        ]
    """
    url = utils.format_urls(base_url, trades.LIST_TRADES)
    params = {
        "contract_hash": contract_hash,
        "pair": pair,
        "from": from_time,
        "to": to_time,
        "limit": limit,
    }
    resp = requests.get(url, params=params)
    return utils.response_else_exception(resp)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Withdrawals allow free movement of tokens from the Switcheo smart contract into your wallet.
This movement of tokens is free of charge.
Once a withdrawal has been executed, tokens in your contract balance would be deducted.
Tokens that are put on hold by orders cannot be withdrawn.
Withdrawals are not instantaneous. Tokens deducted from your contract balance will be added to your wallet balance but
put on hold until the withdrawal has been fully executed.
"""

import logging
import requests
from pyswitcheo import utils
from pyswitcheo.serialization import sign_msg
from pyswitcheo.internal.urls import withdrawals
from pyswitcheo.crypto_utils import (
    encode_msg,
    get_private_key_from_wif,
    get_script_hash_from_wif,
)

logger = logging.getLogger(__name__)


def _create_withdrawal(base_url, priv_key_wif, asset_id, amount, contract_hash, blockchain="NEO"):
    """Creates a withdrawal which can be executed later through execute_withdrawal.

    To be able to make a withdrawal, sufficient funds are required in the contract balance.
    A signature of the request payload has to be provided for this API call.

    Args:
        base_url (str)      : This paramter governs whether to connect to test or mainnet..
        priv_key_wif (str)      : The private key wif of the user.
        asset_id (str)      : The asset symbol or ID to withdraw. for eg. SWTH
        amount (int)        : Amount of tokens to withdraw.
        contract_hash (str) : Switcheo Exchange contract hash to execute the withdraw on.
        blockchain (str)    : Blockchain that the token to withdraw is on. Possible values are: neo.

    Returns:
        An id representing this transaction
        Example response:
        {
          "id": "e0f56e23-2e11-4848-b749-a147c872cbe6"
        }

    """
    signable_params = {
        "blockchain": blockchain,
        "asset_id": asset_id,
        "amount": utils.convert_to_neo_asset_amount(amount, asset_id, base_url),
        "timestamp": utils.get_current_epoch_milli(),
        "contract_hash": contract_hash,
    }
    signable_params_json_str = utils.jsonify(signable_params)

    # The withdrawer's address. Do not include this in the parameters to be signed.
    script_hash = get_script_hash_from_wif(priv_key_wif)

    pk = get_private_key_from_wif(priv_key_wif)
    encoded_msg = encode_msg(signable_params_json_str)
    signature = sign_msg(encoded_msg, pk)

    params = {**signable_params, "signature": signature, "address": script_hash}

    logger.debug("Params being sent to create withdrawal: {0}".format(params))
    url = utils.format_urls(base_url, withdrawals.CREATE_WITHDRAWAL)
    resp = requests.post(url, json=params)
    return utils.response_else_exception(resp)


def _execute_withdrawal(base_url, withdrawal, priv_key_wif):
    """This is the second endpoint required to execute a withdrawal.

    After using the Create Withdrawal endpoint, you will receive a response which requires additional signing.

    Note that a sha256 parameter is provided for convenience to be used directly as part of the ECDSA signature
    process. In production mode, this should be recalculated for additional security.

    Args:
        base_url (str)    : This paramter governs whether to connect to test or mainnet.
        withdrawal (dict) : Withdrawal is the json object returned after executing create_withdrawal end point.
        priv_key_wif (str)    : The private key of the user. (wif)

    Returns:
        If response from the server is HTTP_OK (200) then this returns the requests.response object

    """
    logger.debug("Withdrawal invoke transaction is {0}".format(utils.jsonify(withdrawal)))

    pk = get_private_key_from_wif(priv_key_wif)

    signable_params = {
        "id": withdrawal["id"],
        "timestamp": utils.get_current_epoch_milli()
    }
    signable_params_json_str = utils.jsonify(signable_params)
    encoded_signable_params = encode_msg(signable_params_json_str)
    signature = sign_msg(encoded_signable_params, pk)

    params = {"signature": signature, **signable_params}
    logger.debug("Final params being sent to execute withdrawal {0}".format(params))
    url = utils.format_urls(base_url, withdrawals.EXECUTE_WITHDRAWAL.format(id=withdrawal["id"]))
    resp = requests.post(url, json=params)
    return utils.response_else_exception(resp)



import re
from os import path
from codecs import open  # To use a consistent encoding
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


# Get version without importing, which avoids dependency issues
def get_version():
    with open('pyswitcheo/__init__.py') as version_file:
        return re.search(r"""__version__\s+=\s+(['"])(?P<version>.+?)\1""",
                         version_file.read()).group('version')


install_requires = ['requests==2.19.1', 'neocore', 'jsonschema==2.6.0']


test_requires = ['pytest', 'pytest-sugar', 'pytest-asyncio', 'pytest-cov', 'jsonschema']

extra_requires = {
    'docs': [
        'sphinx >= 1.4',
        'sphinx_rtd_theme',
        'Flask-Sphinx-Themes']}

setup(
    name='pyswitcheo',
    description="Python library to interact with switcheo APIs",
    long_description=long_description,
    version=get_version(),
    include_package_data=True,
    install_requires=install_requires,
    setup_requires=['pytest-runner'],
    tests_require=test_requires,
    extra_requires=extra_requires,
    packages=find_packages(),
    zip_safe=False,
    author="Ankur Srivastava",
    author_email="best.ankur@gmail.com",
    download_url="https://github.com/ansrivas/pyswitcheo/archive/{}.tar.gz".format(get_version()),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"]
)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""Deposits related end points to be used on Switcheo."""

# This endpoint creates a deposit which can be executed through Execute Deposit. To be able to make a deposit,
# sufficient funds are required in the depositing wallet
CREATE_DEPOSIT = "/deposits"


# This is the second endpoint required to execute a deposit.
EXECUTE_DEPOSIT = "/deposits/{id}/broadcast"
